"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const log = require('debug')('bp:worker');
const is_valid_npm_name_1 = __importDefault(require("is-valid-npm-name"));
const fs_1 = __importDefault(require("fs"));
const getParseTime_1 = __importDefault(require("../getParseTime"));
const perf_hooks_1 = require("perf_hooks");
const node_gzip_1 = require("node-gzip");
const core_1 = require("@parcel/core");
const CustomError_1 = require("../errors/CustomError");
const telemetry_utils_1 = __importDefault(require("./telemetry.utils"));
const common_utils_1 = require("./common.utils");
const config_1 = __importDefault(require("../config"));
const p_all_1 = __importDefault(require("p-all"));
function notEmpty(value) {
    return value !== null && value !== undefined;
}
function isDiagnosticError(error) {
    return 'diagnostics' in error;
}
function parseMissingModules(errors) {
    const missingModuleErrors = errors.filter(error => error.message.startsWith('Failed to resolve') &&
        error.origin === '@parcel/core');
    if (!missingModuleErrors.length) {
        return [];
    }
    // There's a better way to get the missing module's name, maybe ?
    const missingModuleRegex = /Failed to resolve '(.+)' from/;
    const missingModules = missingModuleErrors.map(err => {
        const matches = err.message.match(missingModuleRegex);
        if (!matches) {
            throw new CustomError_1.UnexpectedBuildError('Expected to find a file path in the module not found error, but found none. Regex for this might be out of date.');
        }
        const missingFilePath = matches[1];
        let packageNameMatch;
        if (missingFilePath.startsWith('@')) {
            packageNameMatch = missingFilePath.match(/@[^\/]+\/[^\/]+/); // @babel/runtime/object/create -> @babel/runtime
        }
        else {
            packageNameMatch = missingFilePath.match(/[^\/]+/); // babel-runtime/object/create -> babel-runtime
        }
        if (!packageNameMatch) {
            throw new CustomError_1.UnexpectedBuildError('Failed to resolve the missing package name. Regex for this might be out of date.');
        }
        return packageNameMatch[0];
    });
    let uniqueMissingModules = Array.from(new Set(missingModules));
    uniqueMissingModules = uniqueMissingModules.filter(mod => !mod.startsWith(`${uniqueMissingModules[0]}/`));
    return uniqueMissingModules;
}
async function getAssetStats(assets, options) {
    const getAssetStats = async (asset) => {
        if (!asset.file)
            return null;
        const bundleContents = fs_1.default.readFileSync(asset.file, 'utf8');
        let parseTimes = null;
        if (options.calcParse) {
            parseTimes = (0, getParseTime_1.default)(bundleContents);
        }
        const gzipSize = (await (0, node_gzip_1.gzip)(bundleContents)).length;
        const { ext, name } = path_1.default.parse(asset.file);
        return {
            name: name,
            type: ext.slice(1),
            size: asset.size,
            gzip: gzipSize,
            parse: parseTimes,
        };
    };
    const assetStatsPromises = assets
        .filter(asset => !!asset.file)
        .map(getAssetStats);
    const assetStats = (await Promise.all(assetStatsPromises)).filter(notEmpty);
    return assetStats;
}
class BuildUtils {
    static createEntryPoint(packageName, installPath, options) {
        const entryFilename = options.entryFilename || 'index.js';
        const entryPath = path_1.default.join(installPath, entryFilename);
        let importStatement;
        if (options.esm) {
            if (options.customImports) {
                importStatement = `
          import { ${options.customImports.join(', ')} } from '${packageName}'; 
          console.log(${options.customImports.join(', ')})
     `;
            }
            else {
                importStatement = `import p from '${packageName}'; console.log(p)`;
            }
        }
        else {
            if (options.customImports) {
                importStatement = `
        const { ${options.customImports.join(', ')} } = require('${packageName}'); 
        console.log(${options.customImports.join(', ')})
        `;
            }
            else {
                importStatement = `const p = require('${packageName}'); console.log(p)`;
            }
        }
        try {
            fs_1.default.writeFileSync(entryPath, importStatement, 'utf-8');
            return entryPath;
        }
        catch (err) {
            throw new CustomError_1.EntryPointError(err);
        }
    }
    static async compileEntry({ entry, externals, installPath, measureComposition, }) {
        await (0, common_utils_1.updateProjectPeerDependencies)(installPath, Object.fromEntries([...externals.externalBuiltIns, ...externals.externalPackages].map(pack => [pack, '*'])));
        // We use this as a way to pass options to the reporter plugin
        process.env.MEASURE_COMPOSISION = `${measureComposition}`;
        let bundler = new core_1.Parcel({
            entries: [entry],
            mode: 'production',
            env: Object.assign(Object.assign({}, process.env), {
                NODE_ENV: 'production',
            }),
            defaultConfig: '@parcel/config-default',
            shouldAutoInstall: false,
            cacheDir: path_1.default.join(config_1.default.tmp, 'parcel-cache'),
            config: require.resolve(path_1.default.join(__dirname, '..', '..', '.parcelrc')),
            shouldDisableCache: true,
            shouldContentHash: false,
            defaultTargetOptions: {
                sourceMaps: measureComposition,
                outputFormat: 'global',
                shouldOptimize: true,
                isLibrary: false,
                engines: {
                    browsers: [
                        'last 5 Chrome versions',
                        'last 5 Firefox versions',
                        'Safari >= 9',
                        'edge >= 12',
                    ],
                },
            },
        });
        const assets = [];
        let { bundleGraph } = await bundler.run();
        for (let bundle of bundleGraph.getBundles()) {
            assets.push({
                file: bundle.filePath,
                size: bundle.stats.size,
            });
        }
        return { assets };
    }
    static async compilePackage({ name, entry, externals, minifier, installPath, }) {
        const startTime = perf_hooks_1.performance.now();
        let allAssets = [];
        const entries = Object.values(entry);
        const entryPromises = entries.map(entry => async () => {
            const { assets } = await BuildUtils.compileEntry({
                name,
                entry,
                measureComposition: entries.length === 1,
                externals,
                installPath,
            });
            log('Building entry %s', entry);
            allAssets = [...allAssets, ...assets];
        });
        try {
            await (0, p_all_1.default)(entryPromises, { concurrency: 1 });
            telemetry_utils_1.default.compilePackage(name, true, startTime, { minifier });
            return { assets: allAssets };
        }
        catch (err) {
            if (isDiagnosticError(err)) {
                (0, common_utils_1.printDiagnosticError)(err);
            }
            else {
                console.error(err);
            }
            telemetry_utils_1.default.compilePackage(name, false, startTime, { minifier }, err);
            throw err;
        }
    }
    static async buildPackage({ name, installPath, externals, options, }) {
        let entry = {};
        if (options.splitCustomImports) {
            if (!options.customImports || !options.customImports.length) {
                return { assets: [] };
            }
            options.customImports.forEach((importt) => {
                entry[importt] = BuildUtils.createEntryPoint(name, installPath, {
                    customImports: [importt],
                    entryFilename: `${importt}.js`,
                    esm: true,
                });
            });
        }
        else {
            entry['main'] = BuildUtils.createEntryPoint(name, installPath, {
                esm: false,
                customImports: options.customImports,
            });
        }
        log('build start %s', name);
        try {
            const { assets } = await BuildUtils.compilePackage({
                name,
                entry,
                installPath,
                externals,
                debug: options.debug,
                minifier: options.minifier,
            });
            log('build end %s', name);
            log('after compile assets %o', assets);
            const assetStats = await getAssetStats(assets, options);
            return Object.assign({ assets: assetStats || [] }, (options.includeDependencySizes && {
                dependencySizes: require(path_1.default.join(installPath, 'composition.json')),
            }));
        }
        catch (error) {
            if (isDiagnosticError(error)) {
                const missingModules = parseMissingModules(error.diagnostics);
                if (missingModules.length) {
                    if (missingModules.length === 1 && missingModules[0] === name) {
                        throw new CustomError_1.EntryPointError(error.diagnostics);
                    }
                    else {
                        throw new CustomError_1.MissingDependencyError(error.diagnostics, {
                            missingModules,
                        });
                    }
                }
            }
            throw new CustomError_1.BuildError(error);
        }
    }
    static async buildPackageIgnoringMissingDeps({ name, externals, installPath, options }, buildIteration) {
        const buildStartTime = perf_hooks_1.performance.now();
        try {
            const buildResult = await BuildUtils.buildPackage({
                name,
                externals,
                installPath,
                options,
            });
            telemetry_utils_1.default.buildPackage(name, true, buildStartTime, Object.assign(Object.assign({}, options), { buildIteration }));
            return buildResult;
        }
        catch (e) {
            buildIteration++;
            if (e instanceof CustomError_1.MissingDependencyError &&
                e.missingModules.every(mod => (0, is_valid_npm_name_1.default)(mod)) &&
                buildIteration < 4) {
                const { missingModules } = e.extra;
                const newExternals = Object.assign(Object.assign({}, externals), { externalPackages: externals.externalPackages.concat(missingModules) });
                log('%s has missing dependencies, rebuilding without %o', name, missingModules);
                const rebuiltResults = await BuildUtils.buildPackageIgnoringMissingDeps({
                    name,
                    externals: newExternals,
                    installPath,
                    options,
                }, buildIteration);
                telemetry_utils_1.default.buildPackage(name, true, buildStartTime, Object.assign(Object.assign({}, options), { buildIteration,
                    missingModules }));
                return rebuiltResults;
            }
            else {
                telemetry_utils_1.default.buildPackage(name, false, buildStartTime, Object.assign(Object.assign({}, options), { buildIteration }), e);
                throw e;
            }
        }
    }
}
exports.default = BuildUtils;
//# sourceMappingURL=build.utils.js.map